﻿using System;
using System.Collections.Generic;
using System.Text;
using System.Drawing;
using System.Xml.Serialization;
using System.IO;
using System.IO.Ports;

namespace CameraS88
{
    public class ConfigData
    {
        public int MatchingSize = 16;
        public int MatchingThreshold = 32;//12.5%
        public int BlockThreshold = 128;//50%
        //public int MatchingSize = 16;
        public int SelectedCameraIndex = 0;
        public int RotateType = 0;
        public bool UseMonochrome = false;



        public void SaveToFile(String inFilename)
        {

            //XmlSerializerを呼び出す
            XmlSerializer serializer = new XmlSerializer(typeof(ConfigData));
            //ファイルを作る
            FileStream fs = new FileStream(inFilename, FileMode.Create);
            //書き込み
            serializer.Serialize(fs, this);  //sclsはSampleClassのインスタンス名
            //ファイルを閉じる
            fs.Close();

        }

        public bool LoadFromFile(String inFilename)
        {
            bool retVal = false;

            //設定ファイルがあるときのみ読み込みします。
            if (File.Exists(inFilename))
            {
                //XmlSerializerを呼び出す
                XmlSerializer serializer = new XmlSerializer(typeof(ConfigData));
                //ファイルを開く
                FileStream fs = new FileStream(inFilename, FileMode.Open);

                //読み込み
                ConfigData aTempCfg = new ConfigData();
                aTempCfg = (ConfigData)serializer.Deserialize(fs);
                //ファイルを閉じる
                fs.Close();

                //反映
                this.MatchingSize = aTempCfg.MatchingSize;
                this.MatchingThreshold = aTempCfg.MatchingThreshold;
                this.BlockThreshold = aTempCfg.BlockThreshold;
                this.RotateType = aTempCfg.RotateType;
                this.UseMonochrome = aTempCfg.UseMonochrome;
                this.SelectedCameraIndex = aTempCfg.SelectedCameraIndex;

                retVal = true;
            }
            else
            {
                /* 初期化は完了済 */
            }


            return retVal;
        }


    }

    public class CameraS88Item
    {
        public int S88Address;
        public String Name;
        public int X;
        public int Y;
        public int W;
        public int H;
        public int Status;
        public bool Changed;


    }

    public class CameraS88List : List<CameraS88Item>
    {

        public int SelectedS88ItemNo = -1;

        public CameraS88Item Selected()
        {
            if (SelectedS88ItemNo >= 0)
            {
                return this[SelectedS88ItemNo];
            }
            else
            {
                return null;
            }
        
        }

        public void DrawItems(Graphics inCanvas)
        {
            Font aFont = new Font("Arial", 10);

            for (int i = 0; i < this.Count; i++)
            {
                if (i == SelectedS88ItemNo)
                {
                    inCanvas.DrawRectangle(Pens.Red, this[i].X + 1, this[i].Y + 1, this[i].W - 2, this[i].H - 2);
                }
                
                
                if (this[i].Status == 0)
                {

                    inCanvas.DrawRectangle(Pens.Blue, this[i].X, this[i].Y, this[i].W, this[i].H);
                    inCanvas.DrawString(this[i].S88Address.ToString(), aFont, Brushes.Blue, this[i].X, this[i].Y);
                }
                else
                {
                    inCanvas.DrawRectangle(Pens.Yellow, this[i].X, this[i].Y, this[i].W, this[i].H);
                    inCanvas.DrawString(this[i].S88Address.ToString(), aFont, Brushes.Yellow, this[i].X, this[i].Y);


                }

            }
        }

        public int Select(int x, int y)
        {
            int aItemNo = -1;

            for (int i = 0; i < this.Count; i++)
            {
                if( (this[i].X <= x) && (this[i].Y <= y) && (this[i].X + this[i].W > x) && (this[i].Y + this[i].H > y))
                {
                    aItemNo = i;
                    break;
                }

            }

            SelectedS88ItemNo = aItemNo;

            return SelectedS88ItemNo;

        }

        public bool IsSelectedArea(int x, int y)
        {

            if (SelectedS88ItemNo < 0)
            {
                return false;
            }

            if ((this[SelectedS88ItemNo].X <= x) && (this[SelectedS88ItemNo].Y <= y) &&
                (this[SelectedS88ItemNo].X + this[SelectedS88ItemNo].W > x) && 
                (this[SelectedS88ItemNo].Y + this[SelectedS88ItemNo].H > y))
            {
                return true;
            }
            else
            {
                return false;
            }
        }


        public void CheckItems(int inThreshold, int inWidth, int inNSize, ref MatchList outTable)
        {
            int aWCnt = inWidth / inNSize;

            for (int i = 0; i < this.Count; i++)
            {
                int aMatchedCount = 0;
                int aX = (int)Math.Round((double)this[i].X / (double)inNSize);
                int aY = (int)Math.Round((double)this[i].Y / (double)inNSize);
                int aW = (int)Math.Round((double)this[i].W / (double)inNSize);
                int aH = (int)Math.Round((double)this[i].H / (double)inNSize);

                //inCanvas.DrawRectangle(Pens.Blue, this[i].X, this[i].Y, this[i].W, this[i].H);

                for (int y = 0; y < aH; y++)
                {
                    for (int x = 0; x < aW; x++)
                    {
                        int aIndex = (y + aY) * aWCnt + x + aX;

                        if (aIndex < outTable.Count)
                        {

                            if (outTable[aIndex].Matched == true)
                            {
                                aMatchedCount++;
                            }
                        }

                    }

                }

                /* マッチしたカウント数 */
                if (aMatchedCount * 256 / (aW * aH) >= inThreshold)
                {
                    /* 変化チェック */
                    if (this[i].Status == 0)
                    {
                        this[i].Changed = true;
                    }

                    /* 該当している場合 */
                    this[i].Status = 1;
                }
                else
                {
                    /* 変化チェック */
                    if (this[i].Status == 1)
                    {
                        this[i].Changed = true;
                    }

                    this[i].Status = 0;
                }


            }

        }

        public void SaveToFile(String inFilename)
        {

            //XmlSerializerを呼び出す
            XmlSerializer serializer = new XmlSerializer(typeof(CameraS88List));
            //ファイルを作る
            FileStream fs = new FileStream(inFilename, FileMode.Create);
            //書き込み
            serializer.Serialize(fs, this);  //sclsはSampleClassのインスタンス名
            //ファイルを閉じる
            fs.Close();

        }

        public bool LoadFromFile(String inFilename)
        {
            bool retVal = false;
            int i;

            this.Clear();

            //設定ファイルがあるときのみ読み込みします。
            if (File.Exists(inFilename))
            {
                //XmlSerializerを呼び出す
                XmlSerializer serializer = new XmlSerializer(typeof(CameraS88List));
                //ファイルを開く
                FileStream fs = new FileStream(inFilename, FileMode.Open);

                //読み込み
                CameraS88List aTempCfg = new CameraS88List();
                aTempCfg = (CameraS88List)serializer.Deserialize(fs);
                //ファイルを閉じる
                fs.Close();

                //反映
                for (i = 0; i < aTempCfg.Count; i++)
                {
                    this.Add(aTempCfg[i]);
                }

                retVal = true;
            }
            else
            {
                /* 初期化は完了済 */
            }


            return retVal;
        }

    }

    public class MatchItem
    {
        public int Difference;
        public bool Matched;

    }

    public class MatchList : List<MatchItem>
    {

        public void NewSize(int inLength)
        {
            this.Clear();

            for (int i = 0; i < inLength; i++)
            {
                MatchItem aItem = new MatchItem();
                aItem.Difference = 0;
                aItem.Matched = false;

                this.Add(aItem);
            }

        }

        public void FillZero()
        {

            for (int i = 0; i < this.Count; i++)
            {
                this[i].Difference = 0;
                this[i].Matched = false;
            }

        }

    }

    public static class XImages
    {

        public static void ImageProcessing(ref Bitmap ioBitmap, bool inUseMono)
        {
            //モノクロ化
            if (inUseMono == true)
            {
                Monochrome(ref ioBitmap);
            }

            //ノイズ除去(メディアンフィルタ)

            //ヒストグラム均一化
            HistgramNormalization(ref ioBitmap, 10, 250);


        }


        public static void Monochrome(ref Bitmap inBitmap)
        {

            BitmapPlus aBmp = new BitmapPlus(inBitmap);
            aBmp.BeginAccess();

            for (int i = 0; i < inBitmap.Width; i++)
            {
                for (int j = 0; j < inBitmap.Height; j++)
                {
                    // Bitmapの色取得
                    Color bmpCol = aBmp.GetPixel(i, j);

                    int aValue = (bmpCol.R * 76 + bmpCol.G * 150 + bmpCol.B * 29) >> 8;

                    aBmp.SetPixel(i, j, Color.FromArgb(255, aValue, aValue, aValue));
                }
            }
            // Bitmap処理の高速化終了
            aBmp.EndAccess();


        }

        public static void MatchingColor(Bitmap inTemplateImage, Bitmap inTargetImage, int inNSize, ref MatchList outTable)
        {

            BitmapPlus aBmp = new BitmapPlus(inTemplateImage);
            aBmp.BeginAccess();
            BitmapPlus aBmp2 = new BitmapPlus(inTargetImage);
            aBmp2.BeginAccess();

            int aIndex = 0;

            outTable.FillZero();

            for (int ay = 0; ay < inTemplateImage.Height / inNSize; ay++)
            {
                for (int ax = 0; ax < inTemplateImage.Width / inNSize; ax++)
                {

                    int aR = 0;
                    int aG = 0;
                    int aB = 0;

                    int by = ay * inNSize;

                    for (int y = 0; y < inNSize; y++)
                    {
                        int bx = ax * inNSize;

                        for (int x = 0; x < inNSize; x++)
                        {


                            // Bitmapの色取得
                            Color bmpCol = aBmp.GetPixel(bx, by);
                            Color bmpCol2 = aBmp2.GetPixel(bx, by);

                            aR = aR + Math.Abs(bmpCol.R - bmpCol2.R);
                            aG = aG + Math.Abs(bmpCol.G - bmpCol2.G);
                            aB = aB + Math.Abs(bmpCol.B - bmpCol2.B);

                            bx++;
                        }

                        by++;
                    }

                    outTable[aIndex].Difference = (aR + aG + aB) / (3 * 16 * 16);// /(16x16);

                    aIndex++;

                }
            }
            // Bitmap処理の高速化終了
            aBmp.EndAccess();
            aBmp2.EndAccess();

        }

        public static void MatchingMono(Bitmap inTemplateImage, Bitmap inTargetImage, int inNSize, ref MatchList outTable)
        {

            BitmapPlus aBmp = new BitmapPlus(inTemplateImage);
            aBmp.BeginAccess();
            BitmapPlus aBmp2 = new BitmapPlus(inTargetImage);
            aBmp2.BeginAccess();

            int aIndex = 0;

            outTable.FillZero();

            for (int ay = 0; ay < inTemplateImage.Height / inNSize; ay++)
            {
                for (int ax = 0; ax < inTemplateImage.Width / inNSize; ax++)
                {

                    int aR = 0;
                    int by = ay * inNSize;

                    for (int y = 0; y < inNSize; y++)
                    {
                        int bx = ax * inNSize;

                        for (int x = 0; x < inNSize; x++)
                        {
                            // Bitmapの色取得
                            Color bmpCol = aBmp.GetPixel(bx, by);
                            Color bmpCol2 = aBmp2.GetPixel(bx, by);

                            aR = aR + Math.Abs(bmpCol.R - bmpCol2.R);

                            bx++;

                        }

                        by++;
                    }

                    outTable[aIndex].Difference = aR >> 8;// /(16x16);

                    aIndex++;

                }
            }
            // Bitmap処理の高速化終了
            aBmp.EndAccess();
            aBmp2.EndAccess();

        }

        public static void DrawMatchedAreas(Graphics inCanvas, int inThreshold, int inWidth, int inHeight, int inNSize, ref MatchList outTable)
        {

            int aIndex = 0;

            for (int y = 0; y < inHeight / inNSize; y++)
            {
                for (int x = 0; x < inWidth / inNSize; x++)
                {

                    /* 閾値以上のときは描画 */
                    if (outTable[aIndex].Difference >= inThreshold)
                    {
                        inCanvas.DrawRectangle(Pens.SkyBlue, x * inNSize, y * inNSize, inNSize, inNSize);
                        outTable[aIndex].Matched = true;
                    }

                    aIndex++;

                }
            }

        }


        public static void HistgramNormalization(ref Bitmap inBitmap, int inMin, int inMax)
        {
            int aMinR = 255;
            int aMaxR = 0;
            int aMinG = 255;
            int aMaxG = 0;
            int aMinB = 255;
            int aMaxB = 0;

            BitmapPlus aBmp = new BitmapPlus(inBitmap);
            aBmp.BeginAccess();

            /* 最大最小値の探索 */
            for (int i = 0; i < inBitmap.Width; i++)
            {
                for (int j = 0; j < inBitmap.Height; j++)
                {
                    // Bitmapの色取得
                    Color bmpCol = aBmp.GetPixel(i, j);

                    aMaxR = Math.Max(bmpCol.R, aMaxR);
                    aMinR = Math.Min(bmpCol.R, aMinR);
                    aMaxG = Math.Max(bmpCol.G, aMaxG);
                    aMinG = Math.Min(bmpCol.G, aMinG);
                    aMaxB = Math.Max(bmpCol.B, aMaxB);
                    aMinB = Math.Min(bmpCol.B, aMinB);
                }
            }

            aMaxR = Math.Max(inMax, aMaxR);
            aMinR = Math.Min(inMin, aMinR);
            aMaxG = Math.Max(inMax, aMaxG);
            aMinG = Math.Min(inMin, aMinG);
            aMaxB = Math.Max(inMax, aMaxB);
            aMinB = Math.Min(inMin, aMinB);

            /* ヒストグラム均一化処理 */
            for (int i = 0; i < inBitmap.Width; i++)
            {
                for (int j = 0; j < inBitmap.Height; j++)
                {
                    // Bitmapの色取得
                    Color bmpCol = aBmp.GetPixel(i, j);

                    int aR = CalcNormalization(aMinR, aMaxR, bmpCol.R);
                    int aG = CalcNormalization(aMinG, aMaxG, bmpCol.G);
                    int aB = CalcNormalization(aMinB, aMaxB, bmpCol.B);


                    aBmp.SetPixel(i, j, Color.FromArgb(255, aR, aG, aB));
                }
            }

            // Bitmap処理の高速化終了
            aBmp.EndAccess();
        }


        public static int CalcNormalization(int min, int max, int value)
        {

            int aR;

            if (value < min)
            {
                aR = 0;
            }
            else if (value <= max)
            {
                aR = ((value - min) * 255) / (max - min);
            }
            else
            {

                aR = 255;
            }

            return aR;
        }
    }
}
